pub use microrm::prelude::{Database, Entity};
use serde::{Deserialize, Serialize};

use crate::key::KeyType;

// ----------------------------------------------------------------------
// uidc internal types
// ----------------------------------------------------------------------

/// Simple key-value store for persistent configuration
#[derive(Entity)]
pub struct PersistentConfig {
    #[key]
    pub key: String,
    pub value: String,
}

// ----------------------------------------------------------------------
// Session types
// ----------------------------------------------------------------------

#[derive(Entity)]
pub struct Session {
    #[key]
    pub session_id: String,
    pub auth: microrm::RelationMap<SessionAuth>,
    pub expiry: time::OffsetDateTime,
}

#[derive(Entity)]
pub struct SessionAuth {
    pub realm: RealmID,

    pub user: Option<UserID>,

    pub pending_user: Option<UserID>,
    pub pending_challenges: microrm::Serialized<Vec<AuthChallengeType>>,
}

#[derive(Clone, Copy, PartialEq, PartialOrd, Serialize, Deserialize, Debug)]
pub enum AuthChallengeType {
    Username,
    Password,
    TOTP,
    Grid,
    WebAuthn,
}

#[derive(Entity)]
pub struct AuthChallenge {
    #[key]
    pub user_id: UserID,
    #[key]
    pub challenge_type: microrm::Serialized<AuthChallengeType>,
    #[elide]
    pub public: Vec<u8>,
    #[elide]
    pub secret: Vec<u8>,
    pub enabled: bool,
}

// ----------------------------------------------------------------------
// OIDC types
// ----------------------------------------------------------------------

pub struct UserGroupRelation;
impl microrm::Relation for UserGroupRelation {
    type Domain = User;
    type Range = Group;
    const NAME: &'static str = "UserGroup";
}

pub struct GroupRoleRelation;
impl microrm::Relation for GroupRoleRelation {
    type Domain = Group;
    type Range = Role;
    const NAME: &'static str = "GroupRole";
}

#[derive(Clone, Default, Entity)]
pub struct Realm {
    #[key]
    pub shortname: String,

    pub clients: microrm::RelationMap<Client>,
    pub groups: microrm::RelationMap<Group>,
    pub keys: microrm::RelationMap<Key>,
    pub roles: microrm::RelationMap<Role>,
    pub scopes: microrm::RelationMap<Scope>,
    pub users: microrm::RelationMap<User>,
}

#[derive(Entity)]
pub struct Key {
    #[key]
    pub key_id: String,
    pub key_type: microrm::Serialized<KeyType>,
    pub public_data: Vec<u8>,
    pub secret_data: Vec<u8>,
    pub expiry: time::OffsetDateTime,
}

#[derive(Entity)]
pub struct User {
    #[key]
    pub realm: RealmID,
    #[key]
    pub username: String,
    pub auth: microrm::RelationMap<AuthChallenge>,
    pub groups: microrm::RelationDomain<UserGroupRelation>,
}

#[derive(Entity)]
pub struct Group {
    #[key]
    pub realm: RealmID,
    #[key]
    pub shortname: String,
    pub users: microrm::RelationRange<UserGroupRelation>,
    pub roles: microrm::RelationDomain<GroupRoleRelation>,
}

#[derive(Entity)]
pub struct Role {
    #[key]
    pub realm: RealmID,
    /// key publicly-visible name for role
    #[key]
    pub shortname: String,
    pub groups: microrm::RelationRange<GroupRoleRelation>,
}

/// OAuth2 client representation
#[derive(Entity)]
pub struct Client {
    #[key]
    pub realm: RealmID,
    #[key]
    pub shortname: String,
    pub secret: String,
    pub key_type: microrm::Serialized<KeyType>,
    pub redirects: microrm::RelationMap<ClientRedirect>,
    pub scopes: microrm::RelationMap<Scope>,
}

#[derive(Entity)]
pub struct ClientRedirect {
    pub redirect: String,
}

/// Requested group of permissions
#[derive(Entity)]
pub struct Scope {
    #[key]
    pub realm: RealmID,
    #[key]
    pub shortname: String,
    pub roles: microrm::RelationMap<Role>,
}

#[derive(Database)]
pub struct UIDCDatabase {
    pub persistent_config: microrm::IDMap<PersistentConfig>,

    pub realms: microrm::IDMap<Realm>,

    pub sessions: microrm::IDMap<Session>,
}
